#
# estcore.rb
#
#   Copyright (c) 1999 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'nkf'
require 'thread'
require 'delegate'
require 'ftools'
require 'singleton'

require 'amstd/must'
require 'amstd/extmod'
require 'amstd/bug'

require 'tmail/tmail'

require 'est/inetc'
require 'est/readrc'


module EST

  Version = '0.0.5'

  def EST.create
    const_set :CORE, Core.instance
  end


  class CCore

    Export = %w(
      [] mailbox sender popper composer
      notice warn shutdown
      mvmail rmmail mkmail
      add_sender add_popper add_composer add_profile add_refile_rule
      change_profile profile=
      refile )

    include Singleton


    private

    def do_initialize
      confdir = File.expand_path( '~/.est' )
      unless File.exist? confdir then
        File.mkpath confdir
      end
      unless File.directory? confdir then
        puts "cannot create ~/.est (file exists): remove ~/.est and try again"
        exit 1
      end

      require 'est/estconf'
      @corerc = corerc()
      @corerc['confdir'] = confdir
      @corerc['addrfile'] = confdir + '/address'

      @mailbox = Mailset.new
      @sender_all   = {}
      @popper_all   = {}
      @composer_all = {}
      @refile_all   = {}
      @profile_all  = {}
      @sender   = nil
      @popper   = nil
      @composer = nil
      @refile   = nil
      @profile  = nil

      @addrbook = nil

      first_profile = RcReader.read( confdir + '/config' )
      first_profile
    end


    public


    def shutdown
puts 'EST::CORE.shutdown'
      @mailbox.shutdown if @mailbox
puts 'done'
    end


    def notice( str )
      $stderr.puts str
    end

    def warn( str )
      $stderr.puts str
    end
    
    def []( key )
      @corerc[key]
    end


    attr :mailbox
    attr :sender
    attr :popper
    attr :composer


    def add_sender( name, rc )
      @sender_all[ name ] = c = SMTPclient.new( rc )
      @sender ||= c
    end

    def add_popper( name, rc )
      klass = case rc['auth'].downcase
              when 'pop'  then POPclient
              when 'apop' then APOPclient
              else
                bug!
              end
      @popper_all[ name ] = c = klass.new( rc )
      @popper ||= c
    end

    def add_composer( name, rc )
      @composer_all[ name ] = c = Composer.new( rc )
      @composer ||= c
    end

    def add_profile( name, rc )
      @profile_all[ name ] = rc
    end

    def confirm_config( profnm, profiles )
      ent = profiles[ profnm ]
      unless ent['mailbox'] and
             ent['smtp-server'] and
             ent['pop-server'] and
             ent['composer']
        CORE.warn "EST Warning: profile #{profnm} not exist: delete entry"
        profiles.delete profnm
      end
    end

    def change_profile( to )
      unless val = @profile_all[to] then
        CORE.warn "EST Warning: profile #{to} not exist: command canceled."
        return
      end

=begin c
      unless ent = val['mailbox'][to] then
        CORE.warn <<S
EST Warning: non-exist mbox profile '#{to}' was passed:
EST Warning: command canceled.
S
        return
      end
=end c
      @mailbox.inbox = val['mailbox.inbox']
      @mailbox.draft = val['mailbox.draft']
      @mailbox.trash = val['mailbox.trash']

      @sender   = @sender_all[ val['sender'] ]
      @popper   = @popper_all[ val['popper'] ]
      @composer = @composer_all[ val['composer'] ]
    end

    alias profile= change_profile

    def add_refile_rule( name, arr )
      @refile_all[ name ] = arr
      @refile ||= arr
    end

    def mvmail( from, to, idx )
      return unless from = to_mbox_from( from )
      return unless to   = to_mbox_to( to )
      return if from == to
      return unless from[idx]

      to.moveof from, idx
    end

    def rmmail( mb, idx )
      return unless mbox = to_mbox_from( mb )
      mbox.remove idx
    end

    def mkmail( mb, str = nil )
      return nil unless mbox = to_mbox_to( mb )

      port = mbox.new
      port.wopen do |wp|
        if iterator? then
          yield wp
        elsif str then
          wp.write str
        else
          raise ArgumentError, "mkmail: no argument"
        end
      end
      mbox.add port

      return port
    end

    def to_mbox_to( to )
      case to
      when Mailbox then to
      when String  then @mailbox[ to ]
      else
        to.must Mailbox, String
      end
    end

    def to_mbox_from( from )
      case from
      when Mailbox then from
      when String  then @mailbox[ from ]
      else
        to.must Mailbox, String
      end
    end


    def refile
      unless @refile then
        CORE.warn "no refile rule"
        return
      end

      to_mb = nil

      @mailbox.inbox.each_with_index do |port, idx|
        to_mb = nil
        @refile.each do |arr|
          nm, re, to = *arr
          mail = Mail.new( port )

          if re === mail[nm] then
            to_mb = to
            break
          end
        end
        to_mb.mv port if to_mb
      end
    end

  end

  class Core < CCore
  end



  class CMailset

    #
    # export
    #
    #   fetch, [], mbox_names,
    #

    def initialize
      @set = {}
      @rc = {}
      @inbox = nil
      @draft = nil
      @trash = nil
    end

    def shutdown
      @set.each do |name, mb|
        if Mailbox === mb then
          mb.finalize
        end
      end
    end


    def add_mbox( name, rc )
      rc.must Hash
      if @rc[name] then
        if @set[name] then
          raise ArgumentError, "realized mbox '#{name}' is re-registered"
        else
          CORE.warn "mbox '#{name}' registered twice"
        end
      end
# puts "mbox added: name=#{name}"
      @rc[name] = rc
    end

    def del_mbox( name )
      unless @rc[name] then
        raise ArgumentError,
          "not exist mbox '#{name}' was tryed to delete"
      end

      if mb = @set[name] then
        mb.finalize
        @set.delete name
      end
      @rc.delete name
    end


    def []( name )
      name.must String   #D#

      unless ret = @set[ name ] then
        if rc = @rc[ name ] then
          lo = mkloader( rc )
          @set[ name ] = ret = realizembox( name, rc, lo )
        else
          ret = nombox( name )
        end
      end

      ret
    end

    def nombox( name )
      raise ArgumentError, "not exist mbox #{name}"
    end


    class << self
      def sysmb_attr( label )
        label = label.id2name
        module_eval %-
          def #{label}=( name )
            name.must String
            if mb = @rc[ name ] then
              @#{label} = name
            else
              raise ArgumentError,
                "not exist mbox '\#{name}' was given for #{label}"
            end
          end
        -
        module_eval %-
          def #{label}
            self[ @#{label} ]
          end
        -
      end
    end

    sysmb_attr :inbox
    sysmb_attr :draft
    sysmb_attr :trash


    def mbox_names
      ret = @rc.collect {|name,rc| rc['viewname'] || name }
      ret.sort!
      ret
    end


    def mkloader( rc )
      case rc[ 'type' ]
      when 'mh'
        lo = TMail::MhLoader.new( rc['path'] )
      else
        bug! 'mbox type is not MH'
      end

      return lo
    end

    def realizembox( name, rc, lo )
      Mailbox.new( name, rc, lo )
    end

  end

  class Mailset < CMailset
  end



  class CMailbox

    include Enumerable

    def initialize( name, rc, lo )
      name.must String
      @name = name.dup.freeze
      @rc = rc
      @loader = lo
      @array = []
    end

    def finalize
      do_finalize
    end


    private

    def do_initialize
    end

    abstract :do_finalize


    public


    ###  attributes

    attr :name

    def size
      @array.size
    end
    alias length size


    ###  access/search

    def []( idx )
      return nil unless idx
      @array[idx]
    end
    
    def each( &block )
      @array.each( &block )
    end

    def each_index( &block )
      @array.each_index( &block )
    end

    def each_mail
      each{|port| yield Mail.new( port ) }
    end

    def index( port )
      port.must TMail::Port    #D#
      @array.index port
    end

    def rindex( port )
      port.must TMail::Port    #D#
      @array.rindex port
    end

    def indexes( *args )
      @array.indexes *args
    end

    def to_a
      @array.dup
    end


    ###  mail handle

    def add( port )
      do_add port
    end

    def moveof( mbox, idx )
      do_moveof mbox, idx
    end

    def remove( idx )
      do_remove idx
    end

    def new_port
      do_new_port
    end
    alias new new_port
    
    def update
      do_update
    end

    def sort( &block )
      @array.sort!( &block )
    end


    private

    abstract :do_add
    # abstract :do_cp
    abstract :do_move_of
    abstract :do_remove
    abstract :do_new_port

  end

  class Mailbox < CMailbox
  end



  class RealMailbox < Mailbox

    def update
      hash = {}
      each{|port| hash[ port ] = true }
      @loader.each do |p|
        p.unread = true if @name == 'inbox'
        add p unless hash.has_key? p
      end
    end


    private

    def do_initialize
      if entries = load_cache then
        entries.each do |ent|
          port = ent.shift
          add port, ent
        end
      else
puts 'parsing'
        unread = (@name == 'inbox')
        @loader.each do |p|
          p.unread = unread
          add p
        end
      end
    end

    def load_cache
      fn = "#{CORE['confdir']}/cache/#{EST::Interface}/#{@name}"
      entries = nil

      begin
        if File.exist? fn then
          File.open( fn ) do |f|
            break if f.gets.downcase.strip != 'est cache file version 1'
puts 'version ok'

            size = f.gets.strip.to_i
            sign = Marshal.load( f.read size )
            break unless correct_cache? sign
puts 'correct cache'

puts "loading cache: #{@name}"
            entries = Marshal.load( f )
          end
        end
unless entries then
puts "cache miss: #{@name}"
end
      rescue
        $stderr.puts 'error in reading cache file (may not be fatal)'
        $stderr.puts $!.to_s
        entries = nil
      end

      entries
    end

    # abstract correct_cache?()


    def do_finalize
      save_cache if @loader.can_cache?
    end

    def save_cache
puts "saving cache: #{@name}"
      fn = "#{CORE['confdir']}/cache/#{EST::Interface}/#{@name}"

      begin
        File.mkpath File.dirname(fn)
        File.open( fn, 'w' ) do |f|
          f.puts 'est cache file version 1'

          str = Marshal.dump( cache_signature )
          f.puts str.size.to_s
          f.write str

          Marshal.dump( cache_entries, f )
        end
      rescue
        $stderr.puts 'error in writing cache file (may not be fatal)'
        $stderr.puts $!.to_s
        if File.file? fn then
          File.rm_f fn
        end
      end
    end

    # abstract cache_entries()
    # abstract cache_signature()


    def do_add( port )
      port.must TMail::Port

      port.ref
      @array.push port

      port
    end

    def do_moveof( mbox, idx )
      p = @loader.new_port
# puts "do_mv: from=#{port},to=#{p}"
      port = mbox[idx]
      port.mv p
      p.unread = port.unread
      add p
      mbox.remove idx

      return p
    end

    def do_remove( idx )
      return unless port = @array.delete_at( idx )
      port.unref
      return nil
    end

    def do_new_port
      @loader.new_port
    end
  
  end


  class ReferenceError < StandardError ; end


  class VirtualMailbox < Mailbox

    private

    def do_finalize
    end

    def do_add( port )
      port.must TMail::Port
      @array.push port
      port.ref
    end

    def do_move_of( mbox, idx )
      raise ReferenceError, "cannot move mail to virtual mbox"
    end

    def do_remove( idx )
      port = @array.delete_at( idx )
      port.unref if port
    end

    def do_new_port
      raise ReferenceError, "virtual mbox cannot make new mail"
    end

  end



  class Mail < TMail::Mail

    def port
      @port
    end

    def port=( p )
      unless @parsed then
        body
      end
      @port.mv p
      @port = p
    end
      
    def sync
      @port.wopen do |wp|
        wp.write to_s
      end
    end

  end



  class CComposer

    def initialize( rc )
      @rc = rc
    end


    REPLYTO_HEADER = {
      'To'          => %|replyto_to(original)|,
      'Subject'     => %|replyto_subject(original)|,
      'In-Reply-To' => %|original.msgid|,
      'References'  => %|original.msgid|
    }

    def replyto( original )
      case original
      when Mail
        ;
      when TMail::Port
        original = Mail.new( original )
      else
        original.must Mail, TMail::Port
      end

      port = CORE.mkmail( CORE.mailbox.draft ){|wp|
        NEW_HEADER.dup.update( REPLYTO_HEADER ).each do |n,v|
          wp.puts( n + ': ' + eval(v) )
        end
        wp.puts
        replyto_body( wp, original )
      }

      port
    end


    def replyto_to( mail )
      if hed = mail['reply-to']        then return hed.body end
      if hed = mail['from']            then return hed.body end
      if hed = mail['resent-reply-to'] then return hed.body end
      if hed = mail['resent-from']     then return hed.body end
      return ''
    end

    
    def replyto_subject( original )
      sbj = original.subject('')
      sbj = 
        if /\A\s*(?:\[[^\]]+\])?\s*(?:re\:\s*)*(\S.*)\Z/io === sbj then
          'Re: ' + $1
        else
          'Re: ' + sbj
        end

      return sbj
    end

    
    def replyto_body( mail )
      str = <<EOS


In mail "%s"
  %s wrote:

EOS
      str = sprintf( str,
        mail.subject(''), mail.from_phrase('') || mail.from('') )
      mail.body.each {|line| str << '> ' << line }

      return str
    end


    NEW_HEADER = {
      'Date'         => %|DateH.t2s(Time.now)|,
      'Subject'      => %||,
      'From'         => %|@rc['from-address']|,
      'MIME-Version' => %|"1.0"|,
      'Content-Type' => %|"text/plain; charset=iso-2022-jp"|,
      'X-Mailer'     => %|'Mailest version ' + Version|
    }

    def new_mail
      port = CORE.mkmail( CORE.mailbox.draft ){|wp|
        NEW_HEADER.each do |n,v|
          wp << n << ': '
          wp.puts eval(v)
        end
        wp.puts
      }

      port
    end


    def forward( original )
      port = CORE.mkmail( CORE.mailbox.draft ){|wp|
        NEW_HEADER.each do |n,v|
          wp << n << ': '
          wp.puts eval(v)
        end
        if hf = original['to'] then
          if s = hf.addrs.join(', ') then
            wp << 'Resent-To: '
            wp.puts s
          end
        end
        if hf = original['from'] then
          if s = hf.addrs.join(', ') then
            wp << 'Resent-From: '
            wp.puts s
          end
        end
        wp.puts
        wp.puts '---------- forwarded mail:'
        original.body.each do |line|
          wp.puts line
        end
        wp.puts '---------- end forwarded mail'
      }

      port
    end


    def list_items( m )
      [ list_date( m['date'] && m['date'].date ),
        list_from( m ),
        list_subject( m.subject ) ]
    end

    def list_date( ti )
      if ti
        ti.strftime( '%m/%d %H:%M' )
      else
        ''
      end
    end

    def list_from( tm )
      tm.from_phrase(nil) || tm.from('')
    end

    def list_subject( str )
      if /\A\s*\[[^:\]]+:(\d+)\]\s*/o === str then
        $1 + ' ' + $'
      else
        str
      end
    end


    VIEW_HEADER = %w( date from subject x-mailer )

    def view_upper( mail )
      arr = []
      mail.each_header do |k,v|
        arr.push v.inspect unless VIEW_HEADER.include? k
      end
      return arr.join("\n")
    end


    def view_lower( mail )
      tmp = VIEW_HEADER.collect{|k| mail[k].inspect if mail[k] }
      return tmp.compact.join("\n") << "\n\n" << mail_to_viewtext( mail ) 
    end


    def mail_to_viewtext( mail )
      ret = nil
      mtype = mail.maintype.downcase

      case mtype
      when 'text', ''
        ret = mail.body
        case mail.encoding('').downcase
        when '7bit', '8bit', ''
          ;
        when 'base64'
          ret = ret.unpack('m')[0]
        else
          return undecodable_mail( mail )
        end
      when 'multipart'
        mail.parts.each do |m|
          tmp = m.maintype.downcase
          if tmp == 'text' or tmp == '' then
            return mail_to_viewtext(m)
          end
        end
        return unknown_type( mail )
      else
        return unknown_type( mail )
      end

      if NKF.guess( ret ) != NKF::EUC then
        ret = NKF.nkf( '-e', ret )
      end

      # search for "\r\n" from first 200 char
      from = ret.size - 200
      from = from > 0 ? from : 0
      if ret.index( "\r", from ) then
        ret = ret.gsub( "\r\n", "\n" )
      end

      return ret
    end

    def undecodable_mail( mail )
      "cannot decode this mail\n" +
      mail['content-transfer-encoding'].inspect
    end

    def unknown_type( mail )
      "unknown content type\n" +
      mail['content-type'].inspect
    end

    def standby_send( mail )
      mail.to_s
    end

  end

  class Composer < CComposer
  end



  class AddressEntry
    
    def initialize( addr, name, keys = [] )
      keys.must Array
      self.addr = addr
      self.name = name
      @props = keys
    end

    attr :addr

    def addr=( adr )
      @addr = adr.dup
    end

    attr :name

    def name=( nm )
      @name = nm.dup
    end

    def properties
      @props.dup
    end

    def properties=( arr )
      arr.must Array
      arr.each {|i| i.must String }
      @props = arr.dup
    end

    def add_property( prop )
      @props |= prop
    end

    def del_property( prop )
      @props.delete prop
    end

  end


  class AddressBook

    def initialize( fname )
      @filename = File.expand_path( fname )
      @whole = {}
      @prop_tbl = {}
      File.open( @filename ) do |f|
        parse( f )
      end

      @edit = @get_addr = nil
    end


    def parse( f )
      in_group = false

      while line = f.gets do
        line.strip!
        next if line == '%'

        addr = line
        line = f.gets ; line.chomp! ; line.strip! ; name = line
        line = f.gets ; line.chomp! ; line.strip! ; keys = line.split(',')

        ent = AddressEntry.new( addr, name, keys )
        @whole[ addr ] = ent
        keys.each do |k|
          k.strip!
          next if k.size == 0
          unless arr = @prop_tbl[k] then
            @prop_tbl[k] = arr = []
          end
          arr.push ent
        end
      end
    end


    def get_addr( &block )
      unless @get_addr then
        @get_addr = GetAddress.new( self )
      end
      return @get_addr.invoke( block )
    end


    def edit( &block )
      unless @edit then
        @edit = EditAddress.new( self )
      end
      return @edit.invoke( block )
    end

    
    def []( addr )
      return @whole[ addr ]
    end

    def store( addr, entry )
      if @whole.has_key? addr then
        raise ArgumentError, "same address '#{addr}' was registered"
      end
      @whole[addr] = entry
    end

    def each
      @whole.each{|k,ent| yield ent }
    end

    def values
      @whole.values
    end

    def keys
      @whole.keys
    end

    def properties
      a = @prop_tbl.keys
      a.sort!
      a
    end

    def new_property( prop )
      if @prop_tbl.has_key? prop then
        raise ArgumentError, "property '#{prop}' is already registered"
      end
      @prop_tbl[ prop ] = []
    end

    def lookup( keys )
      ret = {}
      keys.each do |k|
        if arr = @prop_tbl[k] then
          arr.each{|a| ret[a] = true }
        end
      end
      return ret.keys
    end

  end


end   # module EST



# post-defines


module TMail

class Port

  def initialize
    @refc = 0
    @unread = false
  end

  attr :unread, true

  def ref
    @refc += 1
  end

  def unref
    @refc -= 1
#puts "port#unref, refc=#{@refc}"
    if @refc < 1 then
#puts "delete file"
      rm
    else
#puts "not delete"
    end
  end

end

class FilePort

  def _dump( limit )
    @filename
  end

  def FilePort._load( str )
    new( str )
  end

end

class Loader

  def can_cache?
    false
  end

end

class MhLoader

  def can_cache?
    true
  end

end

end
